/* =============================================
 * Open Core RayTracer - Animation component
 * Created by Fabio de Albuquerque Dela Antonio
 * fabio914 at gmail.com
 * =============================================
 * This code is licensed under the terms of the
 * GNU General Public License (GPL).
 * =============================================
 * File: Animation.cpp
 * Description: Animation (Uniform B-Spline Curves)
 * =============================================
 */

#include <animation/Animation.h>
#include <geometry/Util.h>

Animation::Animation() {
    startTime = 0.0;
    endTime = 0.0;
    
    currentUp = Vector3D(0.0f, 1.0f, 0.0f);
    currentDirection = Vector3D(1.0f, 0.0f, 0.0f);
    currentRight = Vector3D(0.0f, 0.0f, 1.0f);
    currentPosition = Point3D(0.0f, 0.0f, 0.0f);
}

Animation::Animation(double startTimems, double endTimems) {
    
    if(endTimems < 0 || startTimems < 0)
        fatalError("Animation::Animation", "Invalid times for animation!");
    
    startTime = startTimems;
    endTime = endTimems;
    
    currentUp = Vector3D(0.0f, 1.0f, 0.0f);
    currentDirection = Vector3D(1.0f, 0.0f, 0.0f);
    currentRight = Vector3D(0.0f, 0.0f, 1.0f);
    currentPosition = Point3D(0.0f, 0.0f, 0.0f);
}

Animation::~Animation() {
    for(int i = 0; i < controlPoints.size(); i++)
        delete controlPoints[i];
    
    controlPoints.clear();
}

void Animation::setStart(double startTimems) {
    
    if(startTimems < 0)
        fatalError("Animation::setStart", "Invalid times for animation!");
    
    startTime = startTimems;
}

void Animation::setEnd(double endTimems) {
    
    if(endTimems < 0)
        fatalError("Animation::setEnd", "Invalid times for animation!");
    
    endTime = endTimems;
}

void Animation::addControlPoint(Point3D * point) {
    controlPoints.push_back(point);
}

int Animation::getNumberOfControlPoints() {
    return controlPoints.size();
}

bool Animation::animate(double currentTimems) {
    
    double u;
    
    if(controlPoints.size() < 3)
        error("Animation::animate", "More control points are needed for an animation!");
    else {
        
        if(endTime < startTime)
            fatalError("Animation::animate", "Invalid times for animation!");
            
        if(currentTimems >= startTime && currentTimems <= endTime) {
            
            u = ((currentTimems - startTime)/(endTime - startTime)) * ((double)controlPoints.size() - 3.0) + 3.0;
            
            /* Calculate position, velocity and acceleration */
            q(u);
            ql(u);
            qll(u);
            
            /* Apply correction to up vector, and define current position, direction, right and up vectors */
            currentPosition.setX(position.getX());
            currentPosition.setY(position.getY());
            currentPosition.setZ(position.getZ());
            
            currentUp.setX(0.0f);
            currentUp.setY(1.0f);
            currentUp.setZ(0.0f);
            
            velocity.normalize();
            
            currentDirection.setX(velocity.getX());
            currentDirection.setY(velocity.getY());
            currentDirection.setZ(velocity.getZ());
            
            Vector3D * a = Util::cross_product(velocity, currentUp);
            a->normalize();
            
            currentRight.setX(a->getX());
            currentRight.setY(a->getY());
            currentRight.setZ(a->getZ());
            
            delete a;
            
            Vector3D * b = Util::cross_product(currentRight, velocity);
            b->normalize();
            
            currentUp.setX(b->getX());
            currentUp.setY(b->getY());
            currentUp.setZ(b->getZ());
            
            delete b;
            
            return true;
        }
    }
    
    return false;
}

Point3D * Animation::getCurrentPosition() {
    return new Point3D(currentPosition.getX(), currentPosition.getY(), currentPosition.getZ());
}

Vector3D * Animation::getCurrentUp() {
    return new Vector3D(currentUp.getX(), currentUp.getY(), currentUp.getZ());
}

Vector3D * Animation::getCurrentDirection() {
    return new Vector3D(currentDirection.getX(), currentDirection.getY(), currentDirection.getZ());
}

Vector3D * Animation::getCurrentRight() {
    return new Vector3D(currentRight.getX(), currentRight.getY(), currentRight.getZ());
}

Vector3D * Animation::getCurrentAcceleration() {
    return new Vector3D(acceleration.getX(), acceleration.getY(), acceleration.getZ());
}

void Animation::q(double u) {
    
    register int i = 3; while(u >= i) i++; i--;
    
    Vector3D r0 = Vector3D(controlPoints[i]->getX(), controlPoints[i]->getY(), controlPoints[i]->getZ());
    Vector3D r1 = Vector3D(controlPoints[i - 1]->getX(), controlPoints[i - 1]->getY(), controlPoints[i - 1]->getZ());
    Vector3D r2 = Vector3D(controlPoints[i - 2]->getX(), controlPoints[i - 2]->getY(), controlPoints[i - 2]->getZ());
    Vector3D r3 = Vector3D(controlPoints[i - 3]->getX(), controlPoints[i - 3]->getY(), controlPoints[i - 3]->getZ());
    
    r0.multiply(R0(u - i));
    r1.multiply(R1(u - i));
    r2.multiply(R2(u - i));
    r3.multiply(R3(u - i));
    
    Vector3D sum = Vector3D(0.0f, 0.0f, 0.0f);
    sum.sum(r0);
    sum.sum(r1);
    sum.sum(r2);
    sum.sum(r3);
    
    position = sum;
}

void Animation::ql(double u) {
    
    register int i = 3; while(u >= i) i++; i--;
    
    Vector3D r0 = Vector3D(controlPoints[i]->getX(), controlPoints[i]->getY(), controlPoints[i]->getZ());
    Vector3D r1 = Vector3D(controlPoints[i - 1]->getX(), controlPoints[i - 1]->getY(), controlPoints[i - 1]->getZ());
    Vector3D r2 = Vector3D(controlPoints[i - 2]->getX(), controlPoints[i - 2]->getY(), controlPoints[i - 2]->getZ());
    Vector3D r3 = Vector3D(controlPoints[i - 3]->getX(), controlPoints[i - 3]->getY(), controlPoints[i - 3]->getZ());
    
    r0.multiply(R0l(u - i));
    r1.multiply(R1l(u - i));
    r2.multiply(R2l(u - i));
    r3.multiply(R3l(u - i));
    
    Vector3D sum = Vector3D(0.0f, 0.0f, 0.0f);
    sum.sum(r0);
    sum.sum(r1);
    sum.sum(r2);
    sum.sum(r3);
    
    velocity = sum;
}

void Animation::qll(double u) {
    
    register int i = 3; while(u >= i) i++; i--;
    
    Vector3D r0 = Vector3D(controlPoints[i]->getX(), controlPoints[i]->getY(), controlPoints[i]->getZ());
    Vector3D r1 = Vector3D(controlPoints[i - 1]->getX(), controlPoints[i - 1]->getY(), controlPoints[i - 1]->getZ());
    Vector3D r2 = Vector3D(controlPoints[i - 2]->getX(), controlPoints[i - 2]->getY(), controlPoints[i - 2]->getZ());
    Vector3D r3 = Vector3D(controlPoints[i - 3]->getX(), controlPoints[i - 3]->getY(), controlPoints[i - 3]->getZ());
    
    r0.multiply(R0ll(u - i));
    r1.multiply(R1ll(u - i));
    r2.multiply(R2ll(u - i));
    r3.multiply(R3ll(u - i));
    
    Vector3D sum = Vector3D(0.0f, 0.0f, 0.0f);
    sum.sum(r0);
    sum.sum(r1);
    sum.sum(r2);
    sum.sum(r3);
    
    acceleration = sum;
}
