/* =============================================
 * Open Core RayTracer
 * Created by Fabio de Albuquerque Dela Antonio
 * fabio914 at gmail.com
 *
 * Based on the RayTracer written by
 * Fernando Vieira Paulovich
 * (http://www.icmc.usp.br/~paulovic)
 * =============================================
 * This code is licensed under the terms of the
 * GNU General Public License (GPL).
 * =============================================
 * File: RayTracing.cpp
 * Description: RayTracing methods.
 * =============================================
 */

#include <includes.h>
#include <iostream>
#include <engine/RayTracing.h>
#include <BufferedImage.h>
#include <geometry/Point3D.h>
#include <geometry/Util.h>
#include <geometry/Vector3D.h>
#include <objects/AbstractObject.h>
#include <objects/Color.h>
#include <objects/Light.h>
#include <objects/Material.h>
#include <objects/Scene.h>
#include <engine/Ray.h>
#include <engine/RayShooter.h>

#ifdef VERSOR
#include <vsr/vsr_cga3D_op.h>
#include <vsr/vsr_cga3D_funcs.h>
#endif

#ifdef _multithread_
    #include <pthread.h>
    #define MAX_THREAD 4
#endif

RayTracing::RayTracing() {
    maxsteps = 5;
}

RayTracing::~RayTracing() {

}


BufferedImage * RayTracing::execute(Scene const& scene, Dimension imagesize) {

    //creating the final image
    BufferedImage * image = new BufferedImage(imagesize.width, imagesize.height, 32);

    //processing the viewing parameters
    RayShooter viewing(scene.camera, imagesize);


#ifdef _progress_bar_
    int currPixel = 0;
    int maxPixels = imagesize.height * imagesize.width;
#endif

    for (int y = 0; y < imagesize.height; y++) {
        for (int x = 0; x < imagesize.width; x++) {

            //calculating the ray
            Ray ray = viewing.getRay((float)x, (float)y);

            //tracing the rays
            Color color = trace(scene, ray, 0);

            //set the color
            image->setRGB(x, y, Util::gammacorrect(color).toRGB());

#ifdef _progress_bar_
            currPixel++;
#endif
        }
#ifdef _progress_bar_
        progressBar(currPixel, maxPixels);
#endif
    }


    return image;
}

BufferedImage * RayTracing::executeAntialiasing(Scene const& scene, Dimension imagesize) {

    //creating the final image
    BufferedImage * image = new BufferedImage(imagesize.width, imagesize.height, 32);

    //processing the viewing parameters
    RayShooter viewing(scene.camera, imagesize);

#ifdef _progress_bar_
    int currPixel = 0;
    int maxPixels = imagesize.height * imagesize.width;
#endif

    for (int y = 0; y < imagesize.height; y++) {
        for (int x = 0; x < imagesize.width; x++) {
            //the final color
            Color color;

            for (float ypart = -0.75f; ypart < 1; ypart += 0.5f) {
                for (float xpart = -0.75f; xpart < 1; xpart += 0.5f) {
                    //calculating the ray
                    Ray ray = viewing.getRay(x + xpart, y + ypart);

                    //tracing the rays
                    Color colorpart = trace(scene, ray, 0);

                    //summing-up to the final color
                    color.setRed(color.getRed() + colorpart.getRed());
                    color.setGreen(color.getGreen() + colorpart.getGreen());
                    color.setBlue(color.getBlue() + colorpart.getBlue());

                }
            }

            //compute the average of the anti-aliasing rays
            color.setRed(color.getRed() / 16.0f);
            color.setGreen(color.getGreen() / 16.0f);
            color.setBlue(color.getBlue() / 16.0f);

            //set the color
            image->setRGB(x, y, Util::gammacorrect(color).toRGB());

#ifdef _progress_bar_
            currPixel++;
#endif
        }
#ifdef _progress_bar_
        progressBar(currPixel, maxPixels);
#endif
    }

    return image;
}

#ifdef GAALOP
Color RayTracing::trace(Scene const& scene, Ray const& ray, int step) {
    if (step < maxsteps && ray.getEnergy() > RAYTRACING_MIN_ENERGY) {
        //getting the intersection
        Intersection intersect = scene.intersect(ray);

        if (intersect.valid) {
            //calculating the intersection point
            float * intersection_point = intersect.p;
            Point3D intersection;
            intersection.setX(intersect.p[0]);
            intersection.setY(intersect.p[1]);
            intersection.setZ(intersect.p[2]);

            //calculating the normal
            Vector3D normal = intersect.obj->normal(intersection);

            //calculate the reflected ray direction
            float k = 2 * Util::scalar_product(ray.direction, normal);
            Vector3D refldir(ray.direction.getX() - k * normal.getX(), ray.direction.getY() - k * normal.getY(), ray.direction.getZ() - k * normal.getZ());
            refldir.normalize();

            //calculating the phong contribution

            Color local = phongShadow(intersect.obj, scene, normal, refldir, intersection);
            //calculating the transmitted light
            //...

            Color ret = Color(local.getRed() * ray.getEnergy(), local.getGreen() * ray.getEnergy(), local.getBlue() * ray.getEnergy());

            //std::cout << ret.getRed() << "," << ret.getGreen() << "," << ret.getBlue() << std::endl;

            return ret;

        }
    }

    return scene.background->multiply(ray.getEnergy());
}
#elif VERSOR
Color RayTracing::trace(Scene const& scene, Ray const& ray, int step) {
    if (step < maxsteps && ray.getEnergy() > RAYTRACING_MIN_ENERGY) {
        //getting the intersection
        Intersection intersect = scene.intersect(ray);

        if (intersect.valid) {
            Point3D intersection;
            intersection.setX(intersect.p[0]);
            intersection.setY(intersect.p[1]);
            intersection.setZ(intersect.p[2]);

            //calculating the normal
            Vector3D normal = intersect.obj->normal(intersection);

            //calculate the reflected ray direction
            float k = 2 * Util::scalar_product(ray.direction, normal);
            Vector3D refldir(ray.direction.getX() - k * normal.getX(), ray.direction.getY() - k * normal.getY(), ray.direction.getZ() - k * normal.getZ());
            refldir.normalize();

            //calculating the phong contribution

            Color local = phongShadow(intersect.obj, scene, normal, refldir, intersection);
            //calculating the transmitted light
            //...

            Color ret = Color(local.getRed() * ray.getEnergy(), local.getGreen() * ray.getEnergy(), local.getBlue() * ray.getEnergy());

            //std::cout << ret.getRed() << "," << ret.getGreen() << "," << ret.getBlue() << std::endl;

            return ret;
        }
    }

    return scene.background->multiply(ray.getEnergy());
}
#else
Color RayTracing::trace(Scene const& scene, Ray const& ray, int step) {
    if (step < maxsteps && ray.getEnergy() > RAYTRACING_MIN_ENERGY) {
        //getting the intersection
        Intersection * intersect = scene.intersect(ray);

        if (intersect != NULL) {
            //calculating the intersection point
            Point3D intersection;
            intersection.setX(ray.origin.getX() + intersect->dist * ray.direction.getX());
            intersection.setY(ray.origin.getY() + intersect->dist * ray.direction.getY());
            intersection.setZ(ray.origin.getZ() + intersect->dist * ray.direction.getZ());

            //calculating the normal
            Vector3D normal = intersect->obj->normal(intersection);

            //moving the intersection point on the direction of the normal
            //a small fraction in order to avoid problems related to aritmetic
            //approximation: due to the approximation, the intersection point
            //may be placed inside the object surface, so the next intersection
            //will be the same surface point!
            intersection.setX(intersection.getX() + RAYTRACING_EPSILON * normal.getX());
            intersection.setY(intersection.getY() + RAYTRACING_EPSILON * normal.getY());
            intersection.setZ(intersection.getZ() + RAYTRACING_EPSILON * normal.getZ());

            //calculate the reflected ray direction
            float k = 2 * Util::scalar_product(ray.direction, normal);
            Vector3D refldir(ray.direction.getX() - k * normal.getX(), ray.direction.getY() - k * normal.getY(), ray.direction.getZ() - k * normal.getZ());
            refldir.normalize();

            //calculating the phong contribution
            Color local = phongShadow(intersect->obj, scene, normal, refldir, intersection);

            //calculating the transmitted light
            //...

            //calculating the reflected contribution
            Ray reflray(intersection, refldir, ray.getEnergy() * intersect->obj->material->getREFLT());
            Color reflected = trace(scene, reflray, step + 1);

            Color ret = Color((local.getRed() * ray.getEnergy()) + (reflected.getRed() * reflray.getEnergy()), (local.getGreen() * ray.getEnergy()) + (reflected.getGreen() * reflray.getEnergy()), (local.getBlue() * ray.getEnergy()) + (reflected.getBlue() * reflray.getEnergy()));

            delete intersect;

            return ret;
        }
    }

    return scene.background->multiply(ray.getEnergy());
}
#endif

Color RayTracing::phongShadow(AbstractObject * object, Scene const& scene, Vector3D const& normal, Vector3D const& refldir, Point3D const& intersection) {
    Color color; //final color

    //material of the object
    Material * material = object->material;

    //for each light
    for (int i = 0; i < scene.lights.size(); i++) {
        Light * light = scene.lights[i];

        //adding the ambient contribution
        color.setRed(color.getRed() + material->ambient->getRed());
        color.setGreen(color.getGreen() + material->ambient->getGreen());
        color.setBlue(color.getBlue() + material->ambient->getBlue());

        //shadow ray
        Vector3D aux(intersection, *(light->position));
        aux.normalize();

        Ray shadowray(intersection, aux);
#if defined(GAALOP) || defined(VERSOR)
        Camera * shadow_cam = new Camera();
        shadow_cam->lookat = light->position;
        shadow_cam->eye->setX(intersection.getX() + RAYTRACING_EPSILON * normal.getX());
        shadow_cam->eye->setY(intersection.getY() + RAYTRACING_EPSILON * normal.getY());
        shadow_cam->eye->setZ(intersection.getZ() + RAYTRACING_EPSILON * normal.getZ());
        shadow_cam->updateCam();
        shadowray.camera = shadow_cam;
        Intersection intersect = scene.intersect(shadowray);

        //there is not an object between the light and the object
        if (!intersect.valid) {
#else
        Intersection * intersect = scene.intersect(shadowray);

        //there is not an object between the light and the object
        if (intersect == NULL) {
#endif
            //calculate the light direction
            Vector3D lightray(intersection, *(light->position));
            lightray.normalize();

            float diff = Util::scalar_product(normal, lightray);
            if (diff > 0) {
                //calculating the diffuse component
                color.setRed(color.getRed() + material->diffuse->getRed() * light->diffuse->getRed() * diff);
                color.setGreen(color.getGreen() + material->diffuse->getGreen() * light->diffuse->getGreen() * diff);
                color.setBlue(color.getBlue() + material->diffuse->getBlue() * light->diffuse->getBlue() * diff);

                //calculating the specular component
                float spec = Util::scalar_product(refldir, lightray);
                if (spec > 0) {
                    spec = max(0, (float) pow(spec, material->getNS()));
                    color.setRed(color.getRed() + material->specular->getRed() * light->specular->getRed() * spec);
                    color.setGreen(color.getGreen() + material->specular->getGreen() * light->specular->getGreen() * spec);
                    color.setBlue(color.getBlue() + material->specular->getBlue() * light->specular->getBlue() * spec);
                }
            }
        }
#ifdef LINALG
        if (intersect) {
            delete intersect;
        }
#endif
    }
    return color;
}

Color RayTracing::phong(AbstractObject * object, Scene const& scene, Vector3D normal, Vector3D reflectionray, Point3D intersection) {
    Color color; //final color

    //material of the object
    Material * material = object->material;

    //for each light
    for (int i = 0; i < scene.lights.size(); i++) {
        //calculate the light direction
        Light * light = scene.lights[i];
        Vector3D lightray(intersection, *(light->position));
        lightray.normalize();

        //adding the ambient contribution
        color.setRed(color.getRed() + material->ambient->getRed() * light->ambient->getRed());
        color.setGreen(color.getGreen() + material->ambient->getGreen() * light->ambient->getGreen());
        color.setBlue(color.getBlue() + material->ambient->getBlue() * light->ambient->getBlue());

        float diff = Util::scalar_product(normal, lightray);
        if (diff > 0) {
            //calculating the diffuse component
            color.setRed(color.getRed() + material->diffuse->getRed() * light->diffuse->getRed() * diff);
            color.setGreen(color.getGreen() + material->diffuse->getGreen() * light->diffuse->getGreen() * diff);
            color.setBlue(color.getBlue() + material->diffuse->getBlue() * light->diffuse->getBlue() * diff);

            //calculating the specular component
            float spec = Util::scalar_product(reflectionray, lightray);
            if (spec > 0) {
                spec = max(0, (float) pow(spec, material->getNS()));
                color.setRed(color.getRed() + material->specular->getRed() * light->specular->getRed() * spec);
                color.setGreen(color.getGreen() + material->specular->getGreen() * light->specular->getGreen() * spec);
                color.setBlue(color.getBlue() + material->specular->getBlue() * light->specular->getBlue() * spec);
            }
        }
    }
    return color;
}

