/* =============================================
 * Open Core RayTracer
 * Created by Fabio de Albuquerque Dela Antonio
 * fabio914 at gmail.com
 *
 * Based on the RayTracer written by
 * Fernando Vieira Paulovich
 * (http://www.icmc.usp.br/~paulovic)
 * =============================================
 * This code is licensed under the terms of the
 * GNU General Public License (GPL).
 * =============================================
 * File: main.cpp
 * Description: Main function.
 * =============================================
 */

#include <includes.h>
#include <engine/RayTracing.h>
#include <BufferedImage.h>
#include <geometry/Point3D.h>
#include <geometry/Util.h>
#include <geometry/Vector3D.h>
#include <objects/AbstractObject.h>
#include <objects/Sphere.h>
//#include <objects/Box.h>
#include <objects/Color.h>
#include <objects/Light.h>
#include <objects/Material.h>
#include <objects/Scene.h>
#include <engine/Ray.h>
#include <engine/RayShooter.h>
#include <scenes/SceneLoader.h>

#define TITLE "Open Core RayTracer"
#define VERSION "0.1.7"

extern Surface Display;

void debugInfo(Scene *, Dimension);

/* Flags */
bool output = true;
bool dbg = false;
bool ex_time = false;
bool animate = false;

int main(int argc, char *argv[]) {

    if(argc < 2) {
        printf("Usage: %s <scene description file> [ OPTIONS ]\n", argv[0]);
        printf("Options:\n");
        printf("\t--noout : Result image won't be displayed.\n");
        printf("\t--dbg   : Turn on debug messages.\n");
        printf("\t--time  : Display execution time.\n");
        exit(1);
    }

    printf("%s v.%s\n", TITLE, VERSION);

    SceneLoader sl(argv[1]);

    for(int i = 2; i < argc; i++) {
        if(strcmp(argv[i], "--noout") == 0)
            output = false;
        else if(strcmp(argv[i], "--dbg") == 0)
            dbg = true;
        else if(strcmp(argv[i], "--time") == 0)
            ex_time = true;
        else {
            fprintf(stderr, "Unknown option \"%s\".\n", argv[i]);
            exit(1);
        }
    }

    sl.parseScene();

    Scene * s = sl.getScene();
    Dimension d = sl.getDimension();
    RayTracing rt;

    if(dbg)
        debugInfo(s, d);

    Video::initVideo(d.width, d.height, 32);

    BufferedImage * res;
    time_t start, finish;

    animate = sl.isAnimation();

    if(!animate) {

        if(ex_time)
            start = time(NULL);

        if(sl.getAntialiasing())
            res = rt.executeAntialiasing(*s, d);
        else res = rt.execute(*s, d);

        if(ex_time) {
            finish = time(NULL);
            printf("Execution time: %.2lf s.\n", difftime(finish, start));
        }

        if(output) {
            res->drawImage(Display);
            Video::refreshVideo();
        }

        if(sl.save())
            res->saveBMP(sl.getFilename());

        delete res;

        if(output)
            Video::waitQuit();

    } else {

        double currentTime = 0.0, nFrames = (sl.getFramesPerSecond() * sl.getAnimationTimeLength()/1000.0), deltaTime = (1/sl.getFramesPerSecond())*1000.0;
        unsigned int frameCount = 0;
        char filenames[200];

        if(ex_time)
            start = time(NULL);

        do {

            s->animate(currentTime);

            if(sl.getAntialiasing())
                res = rt.executeAntialiasing(*s, d);
            else res = rt.execute(*s, d);

            if(output) {
                res->drawImage(Display);
                Video::refreshVideo();
            }

            if(sl.save()) {
                sprintf(filenames, "%d_%s", frameCount, sl.getFilename());
                res->saveBMP(filenames);
            }

            delete res;

            if(output)
                if(Video::quitEvent())
                    break;

            currentTime += deltaTime;
            frameCount++;

        } while(frameCount < (unsigned int)nFrames);

        if(ex_time) {
            finish = time(NULL);
            printf("Execution time: %.2lf s.\n", difftime(finish, start));
        }

    }

    Video::finishVideo();

    printf("\n");

    return 0;
}

void debugInfo(Scene * s, Dimension d) {
    printf("Dimension:\n");
    printf("\twidth %d\n", d.width);
    printf("\theight %d\n", d.height);

    printf("Background:\n");
    printf("\tcolor (%f, %f, %f)\n", s->background->getRed(), s->background->getGreen(), s->background->getBlue());

    printf("Camera:\n");
    printf("\teye (%f, %f, %f)\n", s->camera->eye->getX(), s->camera->eye->getY(), s->camera->eye->getZ());
    printf("\tviewup (%f, %f, %f)\n", s->camera->viewup->getX(), s->camera->viewup->getY(), s->camera->viewup->getZ());
    printf("\tlookat (%f, %f, %f)\n", s->camera->lookat->getX(), s->camera->lookat->getY(), s->camera->lookat->getZ());
    printf("\tfov %f\n", s->camera->getFOV());
    printf("\tzvp %f\n", s->camera->getZVP());

    printf("Object count %zd\n", s->objects.size());

    for(int i = 0; i < s->objects.size(); i++) {
        Material * m = s->objects[i]->material;
        printf("Material (%d):\n", i);
        printf("\tambient (%f, %f, %f)\n", m->ambient->getRed(), m->ambient->getGreen(), m->ambient->getBlue());
        printf("\tdiffuse (%f, %f, %f)\n", m->diffuse->getRed(), m->diffuse->getGreen(), m->diffuse->getBlue());
        printf("\tspecular (%f, %f, %f)\n", m->specular->getRed(), m->specular->getGreen(), m->specular->getBlue());
        printf("\tns %f\n", m->getNS());
        printf("\treflt %f\n", m->getREFLT());
        printf("\ttransp %f\n", m->getTRANSP());
    }

    printf("Light count %zd\n", s->lights.size());

    for(int i = 0; i < s->lights.size(); i++) {
        Light * l = s->lights[i];
        printf("Light (%d):\n", i);
        printf("\tambient (%f, %f, %f)\n", l->ambient->getRed(), l->ambient->getGreen(), l->ambient->getBlue());
        printf("\tdiffuse (%f, %f, %f)\n", l->diffuse->getRed(), l->diffuse->getGreen(), l->diffuse->getBlue());
        printf("\tspecular (%f, %f, %f)\n", l->specular->getRed(), l->specular->getGreen(), l->specular->getBlue());
        printf("\tposition (%f, %f, %f)\n", l->position->getX(), l->position->getY(), l->position->getZ());
    }
}

