/* =============================================
 * Open Core RayTracer
 * Created by Fabio de Albuquerque Dela Antonio
 * fabio914 at gmail.com
 *
 * Based on the RayTracer written by
 * Fernando Vieira Paulovich
 * (http://www.icmc.usp.br/~paulovic)
 * =============================================
 * This code is licensed under the terms of the
 * GNU General Public License (GPL).
 * =============================================
 * File: Parser.cpp
 * Description: Parser methods.
 * =============================================
 */

#include <includes.h>
#include <misc/Parser.h>

#define AUX_STRING_SIZE 200 /* string maximum size */

Parser::Parser(char * buffer, int size) {
    Buffer = (char *)malloc(size * sizeof(char));
    if(Buffer == NULL) abort("Could not allocate enough memory.");
    memcpy(Buffer,buffer,size);
    bufferSize = size;
    rewind();
}

Parser::~Parser() {
    free(Buffer);
}

void Parser::rewind() {
    currPos = 0;
    currLine = 0;
    getChar();
    skipWhite();
    skipComment();
    skipWhite();
}

void Parser::message(const char * s) {
    printf("Parser::message (%d): %s\n",currLine,s);
}

void Parser::warning(const char * s) {
    printf("Parser::warning (%d): %s\n",currLine,s);
}

void Parser::error(const char * s) {
    printf("Parser::error (%d): %s\n",currLine,s);
}

void Parser::abort(const char *s) {
    error(s);
    halt();
}

void Parser::expected(const char *s) {
    char ss[AUX_STRING_SIZE];
    sprintf(ss,"%s expected.",s);
    abort(ss);	
}

void Parser::halt() {
    exit(1);
}

char Parser::upCase(char c) {
    if(c>='a' && c<='z') return (c - 32);
    else return c;
}

void Parser::getChar() {
    if(currPos < bufferSize) {
        look = Buffer[currPos];
        if(look == '\n' || look == '\r') currLine++;
        if(look == '\0') look = EOF;
        currPos++;
    } else look = EOF;
}

void Parser::match(char c) {
    if (look != c) {
        char ss[20];
        sprintf(ss,"\'%c\'",c);
        expected(ss);
    } else {
        getChar();
        skipWhite();
        skipComment();
        skipWhite();
    }
}

void Parser::matchNoSkip(char c) {
    if (look != c) {
        char ss[20];
        sprintf(ss,"\'%c\'",c);
        expected(ss);
    } else getChar();
}

bool Parser::isAlpha(char c) {	    
    return ((c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z'));
}

bool Parser::isDigit(char c) {
    return (c >= '0' && c <= '9');
}

bool Parser::isAlNum(char c) {
    return (isAlpha( c ) || isDigit( c ) || c=='.' || c=='-' || c=='_');
}

bool Parser::isWhite(char c) {
    return (c==' ' || c=='\n' || c=='\r');
}

void Parser::skipWhite(void) {
    while (isWhite(look)) getChar();
}

void Parser::skipUntil(char c) {
    while (look != c) {
        getChar();
        if (look == EOF) match( c );
    }
    match( c );
    skipWhite();
}

void Parser::skipUntilEnd(void) {
    while (look != '\n' && look != '\r') {
        getChar();
        if (look == EOF) break;
    }
    skipWhite();
}

char * Parser::getItem(void) {
    char *Token;
    Token = (char *)calloc(AUX_STRING_SIZE,sizeof(char));
    if(Token == NULL) abort("Could not allocate enough memory.");
    int i=0;
    if (!isAlNum(look)) expected("Word");
    while (isAlNum(look)) {
        Token[i] = look;
        getChar();
        i++;
        if (i == AUX_STRING_SIZE-1) expected("Less chars");
    }
    Token[i] = '\0';
    skipWhite();
    skipComment();
    skipWhite();
    return Token;
}

char * Parser::getString(void) {
    char *String;
    String = (char *)calloc(AUX_STRING_SIZE,sizeof(char));
    if(String == NULL) abort("Could not allocate enough memory.");
    matchNoSkip('"');
    int i = 0;
    while (look != '"') {
        String[i] = look;
        getChar();
        i++;
        if(i == AUX_STRING_SIZE - 1) abort("String is too large");
    }
    String[i] = '\0';
    match('"');
    return String;
}

int Parser::getNumInt(void) {
    char * numero = getNum();
    if(isInt(numero)) {
        int ret = atoi(numero);
        free(numero);
        return ret;
    } else expected("Integer");
    return -1;
}

float Parser::getNumFloat(void) {
    char * flt = getNumF();
    if(isFloat(flt)) {
        float ret = (float) atof(flt);
        free(flt);
        return ret;
    } else expected("Float");
    return -1;
}

char * Parser::getNum(void) {
    char * Value;
    Value = (char *)calloc(AUX_STRING_SIZE,sizeof(char));
    if(Value == NULL) abort("Could not allocate enough memory.");
    int i=0;
    if (!isDigit(look) && look!='-') expected("Number");
    while (isDigit(look) || look == '-') {
        Value[i] = look;
        getChar();
        i++;
        if (i == AUX_STRING_SIZE-1) expected("Less chars");
    }
    Value[i] = '\0';
    skipWhite();
    skipComment();
    skipWhite();
    return Value;
}

char * Parser::getNumF(void) {
    char * Value;
    Value = (char *)calloc(AUX_STRING_SIZE,sizeof(char));
    if(Value == NULL) abort("Could not allocate enough memory.");
    int i=0;
    if (!isDigit(look) && look!='-' && look != '.') expected("Number");
    while (isDigit(look) || look == '-' || look == '.') {
        Value[i] = look;
        getChar();
        i++;
        if (i == AUX_STRING_SIZE-1) expected("Less chars");
    }
    Value[i] = '\0';
    skipWhite();
    skipComment();
    skipWhite();
    return Value;
}


bool Parser::isFloat(char *s) {
    for (unsigned int i=0;i < strlen(s);i++)
        if (!isDigit(s[i]) && s[i]!='-' && s[i]!='.') {
            expected("Float");
            return false;
        }
    return true;
}

bool Parser::isInt(char *s) {
    for (unsigned int i=0;i < strlen(s);i++)
        if (!isDigit(s[i]) && s[i]!='-') {
            expected("Integer");
            return false;
        }
    return true;
}

bool Parser::isEnd() {
    return (look == EOF);
}

void Parser::strToUpper(char * s) {
    register unsigned int i = 0;
    while (i < strlen(s)) {
        s[i] = upCase(s[i]);
        i++;
    }	
}

char Parser::getLook() {
    return look;
}

void Parser::skipComment() {
    if(look == '/') {
        getChar();
        if(look == '/') {
            skipUntilEnd();
            if(look == '/')
                skipComment();        
        } else expected("Comment");
    }
}
