/* =============================================
 * Open Core RayTracer
 * Created by Fabio de Albuquerque Dela Antonio
 * fabio914 at gmail.com
 *
 * Based on the RayTracer written by
 * Fernando Vieira Paulovich
 * (http://www.icmc.usp.br/~paulovic)
 * =============================================
 * This code is licensed under the terms of the
 * GNU General Public License (GPL).
 * =============================================
 * File: Scene.cpp
 * Description: Scene methods.
 * =============================================
 */

#include <includes.h>
#include <objects/Scene.h>
#include <engine/Ray.h>
#include <objects/AbstractObject.h>
#include <objects/Camera.h>
#include <objects/Light.h>

Scene::Scene() {
    camera = new Camera();
    camera->updateCam();
    background = new Color();
}

Scene::~Scene() {
    delete camera;
    delete background;

    for(int i = 0; i < objects.size(); i++)
        delete objects[i];

    for(int i = 0; i < lights.size(); i++)
        delete lights[i];

    objects.clear();
    lights.clear();
}

#if defined(GAALOP) || defined(VERSOR)
Intersection Scene::intersect(Ray const& ray) const {
    float mindist = POSITIVE_INFINITY;
    Intersection nearest;
    for (auto const& obj: objects) {
        Intersection intersection = obj->intersect(ray);
        if (!intersection.valid)
            continue;
        float dist = intersection.distTo(*this->camera);
        if (dist <= mindist) {
            mindist = dist;
            nearest = intersection;
        }
    }
    return nearest;
}
#else
Intersection * Scene::intersect(Ray const& ray) const {
    float mindist = POSITIVE_INFINITY;
    Intersection * pair = new Intersection(); //intersection pair

    for (int i = 0; i < objects.size(); i++) {
        float dist = objects[i]->intersect(ray);

        if (dist >= 0 && mindist > dist) {
            mindist = dist;
            pair->obj = objects[i];
            pair->dist = dist;
        }
    }

    if(pair->obj == NULL) {
        delete pair;
        return NULL;
    } else return pair;
}
#endif

#ifdef GAALOP
float Intersection::distTo(Camera const& cam) const
{
#pragma gpc begin
    VPlane = mv_from_array(cam.visibility_plane, e1, e2, e3, einf);
    Point = mv_from_array(p, e1, e2, e3, einf, e0);
#pragma clucalc begin
    ?dist = Point . VPlane;
#pragma clucalc end
    float d = mv_get_bladecoeff(dist, 1);
#pragma gpc end
    return d;
}
#elif VERSOR
float Intersection::distTo(Camera const& cam) const
{
    vsr::cga3D::Scalar dist = p <= cam.visibility_plane;
    return dist[0];
}
#endif

void Scene::animate(double currentTimems) {
    for(int i = 0; i < objects.size(); i++)
        objects[i]->animate(currentTimems);
}


