/* =============================================
 * Open Core RayTracer
 * Created by Fabio de Albuquerque Dela Antonio
 * fabio914 at gmail.com
 *
 * Based on the RayTracer written by
 * Fernando Vieira Paulovich
 * (http://www.icmc.usp.br/~paulovic)
 * =============================================
 * This code is licensed under the terms of the
 * GNU General Public License (GPL).
 * =============================================
 * File: Sphere.cpp
 * Description: Sphere methods.
 * =============================================
 */

#include <includes.h>
#include <iostream>
#include <objects/Sphere.h>
#include <engine/Ray.h>
#include <geometry/Point3D.h>
#include <geometry/Vector3D.h>

Sphere::Sphere(Point3D Center, float Radius) {
    position = Point3D(Center.getX(), Center.getY(), Center.getZ());
    radius = Radius;
    if(radius == 0.0f) radius = 0.01f;
}

void Sphere::update() {
#if defined(GAALOP) || defined(VERSOR)
    float Cx = position.getX();
    float Cy = position.getY();
    float Cz = position.getZ();
#endif
#ifdef GAALOP
#pragma gpc begin
#pragma clucalc begin
    ?S = *SphereN3(Cx, Cy, Cz, radius);
#pragma clucalc end
    sphere = mv_to_array(S, e1, e2, e3, einf, e0);
#pragma gpc end
#elif VERSOR
    sphere = vsr::Ro::sphere(Cx, Cy, Cz, radius);
#endif
}

Sphere::~Sphere() {
}

#if defined(GAALOP) || defined(VERSOR)
Intersection
#else
float
#endif
Sphere::intersect(Ray const& ray) {
#ifdef GAALOP
    Camera * cam = ray.camera;
    float visibility_plane[4];
    std::copy(cam->visibility_plane, cam->visibility_plane+4, visibility_plane);
#pragma gpc begin
    VPlane = mv_from_array(visibility_plane, e1, e2, e3, einf);
    Sphere = mv_from_array(sphere, e1, e2, e3, einf, e0);
    Ray = mv_from_array(ray.ray, e1^e2, e1^e3, e1^einf, e2^e3, e2^einf, e3^einf);
#pragma clucalc begin
    ?PP = *(Ray ^ Sphere);
    ?hasIntersection = PP . PP;
#pragma clucalc end
    if (hasIntersection <= 0)
        return Intersection();
#pragma clucalc begin
    denom = e0 . PP;
    P1 = (sqrt(PP . PP) + PP) / denom;
    ?P1normed = P1 / (-P1.einf); // scalar value of e0
    P2 = (-sqrt(PP . PP) + PP) / denom;
    ?P2normed = P2 / (-P2.einf); // scalar value of e0
    ?d1 = P1normed . VPlane; // TODO fixme
    ?d2 = P2normed . VPlane;
#pragma clucalc end
    float t1 = mv_get_bladecoeff(d1, 1);
    float t2 = mv_get_bladecoeff(d2, 1);
    bool P1V = t1 > 0;
    bool P2V = t2 > 0;
    if (!P1V && !P2V)
    {
        return Intersection();
    }
    Intersection i;
    i.valid = true;
    i.obj = this;
    float * p = i.p;
    if (P1V && !P2V)
    {
        p = mv_to_array(P1normed, e1, e2, e3, einf, e0);
        return i;
    }
    if (P2V && !P1V)
    {
        p = mv_to_array(P2normed, e1, e2, e3, einf, e0);
        return i;
    }
    if (P1V && P2V)
    {
        if (t1 < t2)
        {
            p = mv_to_array(P1normed, e1, e2, e3, einf, e0);
        }
        else
        {
            p = mv_to_array(P2normed, e1, e2, e3, einf, e0);
        }
        return i;
    }
#pragma gpc end
#elif VERSOR 
    vsr::cga3D::Pair pp = (ray.ray ^ sphere).dual();
    bool hasIntersection = (pp <= pp)[0] > 0;
    if (!hasIntersection)
        return Intersection();
    std::vector<vsr::cga3D::Point> points = vsr::Ro::split(pp);
    vsr::cga3D::Point p1 = points[0];
    vsr::cga3D::Point p2 = points[1];
    float dist1 = (p1 <= ray.camera->visibility_plane)[0];
    float dist2 = (p1 <= ray.camera->visibility_plane)[0];

    bool P1V = dist1 > 0;
    bool P2V = dist2 > 0;
    if (!P1V && !P2V)
    {
        return Intersection();
    }
    Intersection i;
    i.valid = true;
    i.obj = this;
    if (P1V && !P2V)
    {
        i.p = p1;
        return i;
    }
    if (P2V && !P1V)
    {
        i.p = p2;
        return i;
    }
    if (P1V && P2V)
    {
        if (dist1 < dist2)
            i.p = p1;
        else
            i.p = p2;
        return i;
    }
#else
    //translate ray origin in order to put the center sphere on the system
    //origin (0,0,0)
    float x = ray.origin.getX() - position.getX();
    float y = ray.origin.getY() - position.getY();
    float z = ray.origin.getZ() - position.getZ();

    float d = x * ray.direction.getX() + y * ray.direction.getY() + z * ray.direction.getZ();
    float t = d * d - x * x - y * y - z * z + radius * radius;

    if (t < 0) { //no intersection
      return -1;
    }

    float s = -d - (float) sqrt(t);
    if (s > 0) { //distance from the origin on the given direction
      return s;
    }

    s = -d + (float) sqrt(t);
    if (s > 0) { //distance from the origin on the given direction
      return s;
    }

    return -1; //no intersection
#endif
}

Vector3D Sphere::normal(Point3D intersection) {
    return Vector3D((intersection.getX() - position.getX()) / radius, (intersection.getY() - position.getY()) / radius, (intersection.getZ() - position.getZ()) / radius);
}

float Sphere::getRadius() { return radius; }
void Sphere::setRadius(float Radius) { radius = Radius; if(radius == 0.0f) radius = 0.01f; update(); }
