/* =============================================
 * Open Core RayTracer
 * Created by Fabio de Albuquerque Dela Antonio
 * fabio914 at gmail.com
 *
 * Based on the RayTracer written by
 * Fernando Vieira Paulovich
 * (http://www.icmc.usp.br/~paulovic)
 * =============================================
 * This code is licensed under the terms of the
 * GNU General Public License (GPL).
 * =============================================
 * File: SceneLoader.cpp
 * Description: SceneLoader methods.
 * =============================================
 */

#include <includes.h>
#include <scenes/SceneLoader.h>
//#include <objects/Box.h>
#include <objects/Camera.h>
#include <objects/Color.h>
#include <objects/Light.h>
#include <objects/Material.h>
#include <objects/Scene.h>
#include <objects/Sphere.h>
#include <engine/RayShooter.h>
#include <misc/Parser.h>

#define SDF_SIZE 100000 /* Scene Description File size */

SceneLoader::SceneLoader() {
    initLoader("scene.txt");
}

SceneLoader::SceneLoader(char * filename) {
    initLoader(filename);
}

SceneLoader::~SceneLoader() {
    delete p;
    free(buffer);
}

void SceneLoader::initLoader(char const* filename) {
    FILE * in;
    if((in = fopen(filename, "r")) == NULL) {
        fatalError("SceneLoader::initLoader", "Could not open scene file.");
    }

    buffer = (char *)malloc(SDF_SIZE * sizeof(char));

    if(buffer == NULL)
        fatalError("SceneLoader::initLoader","Could not allocate enough memory.");

    int i;

    for(i = 0; i < SDF_SIZE; i++)
        buffer[i] = '\0';

    i = 0;
    while(!feof(in)) {
        buffer[i] = fgetc(in);
        i++;
        if(i == SDF_SIZE)
            fatalError("SceneLoader::initLoader","Input file is too big.");
    }

    s = new Scene();
    p = new Parser(buffer, SDF_SIZE);
    antialiasing = false;
    filename = NULL;
    saveImage = false;
    animatedScene = false;
}

float SceneLoader::parseFloat(void) {
    return p->getNumFloat();
}

void SceneLoader::parseArray(float *x, float *y, float *z) {
    p->match('(');
    (*x) = p->getNumFloat();
    p->match(',');
    (*y) = p->getNumFloat();
    p->match(',');
    (*z) = p->getNumFloat();
    p->match(')');
}

Material * SceneLoader::parseMaterial(void) {
    Material * ret = new Material();
    float r, g, b;
    char aux[200];

    p->match('{');
    while(p->getLook() != '}') {
        char * token = p->getItem();

        if(strcmp(token, "ns") == 0) {
            p->match('=');
            ret->setNS(parseFloat());
            p->match(';');
        }

        else if(strcmp(token, "reflt") == 0) {
            p->match('=');
            ret->setREFLT(parseFloat());
            p->match(';');
        }

        else if(strcmp(token, "transp") == 0) {
            p->match('=');
            ret->setTRANSP(parseFloat());
            p->match(';');
        }

        else if(strcmp(token, "ambient") == 0) {
            p->match('=');
            parseArray(&r, &g, &b);
            ret->setAmbient(r, g, b);
            p->match(';');
        }

        else if(strcmp(token, "diffuse") == 0) {
            p->match('=');
            parseArray(&r, &g, &b);
            ret->setDiffuse(r, g, b);
            p->match(';');
        }

        else if(strcmp(token, "specular") == 0) {
            p->match('=');
            parseArray(&r, &g, &b);
            ret->setSpecular(r, g, b);
            p->match(';');
        }

        else {
            sprintf(aux, "Unknown token \'%s\'", token);
            fatalError("SceneLoader::parseMaterial", aux);
        }

        free(token);
    }
    p->match('}');
    return ret;
}

/*
Box * SceneLoader::parseBox(void) {
    Box * ret = new Box();
    float x, y, z;
    char aux[200];

    p->match('{');
    while(p->getLook() != '}') {
        char * token = p->getItem();

        if(strcmp(token, "size") == 0) {
            p->match('=');
            parseArray(&x, &y, &z);
            ret->setSX(x);
            ret->setSY(y);
            ret->setSZ(z);
            p->match(';');
        }

        else if(strcmp(token, "position") == 0) {
            p->match('=');
            parseArray(&x, &y, &z);
            ret->position.setX(x);
            ret->position.setY(y);
            ret->position.setZ(z);
            p->match(';');
        }

        else if(strcmp(token, "material") == 0) {
            ret->material = parseMaterial();
        }

        else if(strcmp(token, "animation") == 0) {
            if(ret->animation != NULL)
                delete ret->animation;
            ret->animation = parseAnimation();
        }

        else {
            sprintf(aux, "Unknown token \'%s\'", token);
            fatalError("SceneLoader::parseBox", aux);
        }

        free(token);
    }
    p->match('}');
    return ret;
}
*/

Sphere * SceneLoader::parseSphere(void) {
    Sphere * ret = new Sphere();
    float x, y, z;
    char aux[200];

    p->match('{');
    while(p->getLook() != '}') {
        char * token = p->getItem();

        if(strcmp(token, "radius") == 0) {
            p->match('=');
            ret->setRadius(parseFloat());
            p->match(';');
        }

        else if(strcmp(token, "position") == 0) {
            p->match('=');
            parseArray(&x, &y, &z);
            ret->position.setX(x);
            ret->position.setY(y);
            ret->position.setZ(z);
            p->match(';');
        }

        else if(strcmp(token, "material") == 0) {
            ret->material = parseMaterial();
        }

        else if(strcmp(token, "animation") == 0) {
            if(ret->animation != NULL)
                delete ret->animation;
            ret->animation = parseAnimation();
        }

        else {
            sprintf(aux, "Unknown token \'%s\'", token);
            fatalError("SceneLoader::parseSphere", aux);
        }

        free(token);
    }
    p->match('}');
    ret->update();
    return ret;
}

Light * SceneLoader::parseLight(void) {
    Light * ret = new Light();
    float r, g, b;
    float x, y, z;
    char aux[200];

    p->match('{');
    while(p->getLook() != '}') {
        char * token = p->getItem();

        if(strcmp(token, "ambient") == 0) {
            p->match('=');
            parseArray(&r, &g, &b);
            ret->ambient->setRed(r);
            ret->ambient->setGreen(g);
            ret->ambient->setBlue(b);
            p->match(';');
        }

        else if(strcmp(token, "diffuse") == 0) {
            p->match('=');
            parseArray(&r, &g, &b);
            ret->diffuse->setRed(r);
            ret->diffuse->setGreen(g);
            ret->diffuse->setBlue(b);
            p->match(';');
        }

        else if(strcmp(token, "specular") == 0) {
            p->match('=');
            parseArray(&r, &g, &b);
            ret->specular->setRed(r);
            ret->specular->setGreen(g);
            ret->specular->setBlue(b);
            p->match(';');
        }

        else if(strcmp(token, "position") == 0) {
            p->match('=');
            parseArray(&x, &y, &z);
            ret->position->setX(x);
            ret->position->setY(y);
            ret->position->setZ(z);
            p->match(';');
        }

        else {
            sprintf(aux, "Unknown token \'%s\'", token);
            fatalError("SceneLoader::parseLight", aux);
        }

        free(token);
    }
    p->match('}');
    return ret;
}

Camera * SceneLoader::parseCamera(void) {
    Camera * ret = new Camera();
    float x, y, z;
    char aux[200];

    p->match('{');
    while(p->getLook() != '}') {
        char * token = p->getItem();

        if(strcmp(token, "eye") == 0) {
            p->match('=');
            parseArray(&x, &y, &z);
            ret->eye->setX(x);
            ret->eye->setY(y);
            ret->eye->setZ(z);
            p->match(';');
        }

        else if(strcmp(token, "lookat") == 0) {
            p->match('=');
            parseArray(&x, &y, &z);
            ret->lookat->setX(x);
            ret->lookat->setY(y);
            ret->lookat->setZ(z);
            p->match(';');
        }

        else if(strcmp(token, "viewup") == 0) {
            p->match('=');
            parseArray(&x, &y, &z);
            ret->viewup->setX(x);
            ret->viewup->setY(y);
            ret->viewup->setZ(z);
            p->match(';');
        }

        else if(strcmp(token, "fov") == 0) {
            p->match('=');
            ret->setFOV(parseFloat());
            p->match(';');
        }

        else if(strcmp(token, "zvp") == 0) {
            p->match('=');
            ret->setZVP(parseFloat());
            p->match(';');
        }

        else {
            sprintf(aux, "Unknown token \'%s\'", token);
            fatalError("SceneLoader::parseCamera", aux);
        }

        free(token);
    }
    p->match('}');
    ret->updateCam();
    return ret;
}

Dimension SceneLoader::parseDimension(void) {
    Dimension ret;
    char aux[200];

    p->match('{');
    while(p->getLook() != '}') {
        char * token = p->getItem();

        if(strcmp(token, "width") == 0) {
            p->match('=');
            ret.width = p->getNumInt();
            p->match(';');
        }

        else if(strcmp(token, "height") == 0) {
            p->match('=');
            ret.height = p->getNumInt();
            p->match(';');
        }

        else {
            sprintf(aux, "Unknown token \'%s\'", token);
            fatalError("SceneLoader::parseDimension", aux);
        }

        free(token);
    }
    p->match('}');
    return ret;
}

Color SceneLoader::parseBackground(void) {
    Color ret;
    float r, g, b;
    char aux[200];

    p->match('{');
    while(p->getLook() != '}') {
        char * token = p->getItem();

        if(strcmp(token, "color") == 0) {
            p->match('=');
            parseArray(&r, &g, &b);
            ret.setRed(r);
            ret.setGreen(g);
            ret.setBlue(b);
            p->match(';');
        }

        else {
            sprintf(aux, "Unknown token \'%s\'", token);
            fatalError("SceneLoader::parseDimension", aux);
        }

        free(token);
    }
    p->match('}');
    return ret;
}

Animation * SceneLoader::parseAnimation() {
    Animation * animation = new Animation();
    float x, y, z;
    Point3D * auxp;
    char aux[200];

    p->match('{');
    while(p->getLook() != '}') {
        char * token = p->getItem();

        if(strcmp(token, "start") == 0) {
            p->match('=');
            x = p->getNumFloat();
            p->match(';');
            animation->setStart(x);
        }

        else if(strcmp(token, "end") == 0) {
            p->match('=');
            x = p->getNumFloat();
            p->match(';');
            animation->setEnd(x);
        }

        else if(strcmp(token, "points") == 0) {
            p->match('=');
            p->match('{');
            while(1) {
                parseArray(&x, &y, &z);
                auxp = new Point3D(x, y, z);
                animation->addControlPoint(auxp);

                if(p->getLook() == ',') {
                    p->match(',');
                    continue;
                } else break;
            }
            p->match('}');
            p->match(';');
        }

        else {
            sprintf(aux, "Unknown token \'%s\'", token);
            fatalError("SceneLoader::parseAnimation", aux);
        }

        free(token);
    }
    p->match('}');

    return animation;
}

void SceneLoader::parseScene(void) {
    char aux[200];

    char * start = p->getItem();

    if(strcmp(start, "scene") != 0) {
        sprintf(aux, "Unknown token \'%s\'", start);
        fatalError("SceneLoader::parseScene", aux);
    }

    free(start);

    p->match('{');
    while(p->getLook() != '}') {
        char * token = p->getItem();

        if(strcmp(token, "antialiasing") == 0) {
            antialiasing = true;
            p->match(';');
        }

        else if(strcmp(token, "background") == 0) {
            *(s->background) = parseBackground();
        }

        else if(strcmp(token, "dimension") == 0) {
            d = parseDimension();
        }

        else if(strcmp(token, "camera") == 0) {
            delete s->camera;
            s->camera = parseCamera();
        }

        else if(strcmp(token, "box") == 0) {
            //s->objects.push_back(parseBox());
        }

        else if(strcmp(token, "sphere") == 0) {
            s->objects.push_back(parseSphere());
        }

        else if(strcmp(token, "light") == 0) {
            s->lights.push_back(parseLight());
        }

        else if(strcmp(token, "save") == 0) {
            p->match('(');
            filename = p->getString();
            p->match(')');
            p->match(';');
            saveImage = true;
        }

        else if(strcmp(token, "animation") == 0) {
            p->match('(');
            fps = p->getNumFloat();
            p->match(',');
            totalAnimationTime = p->getNumFloat();
            p->match(')');
            p->match(';');
            animatedScene = true;
        }

        else {
            sprintf(aux, "Unknown token \'%s\'", token);
            fatalError("SceneLoader::parseScene", aux);
        }

        free(token);
    }
    p->match('}');
}

Scene * SceneLoader::getScene(void) { return s; }
Dimension SceneLoader::getDimension(void) { return d; }
bool SceneLoader::getAntialiasing(void) { return antialiasing; }
char * SceneLoader::getFilename(void) { return filename; }
bool SceneLoader::save(void) { return saveImage; }
bool SceneLoader::isAnimation(void) { return animatedScene; }
float SceneLoader::getFramesPerSecond() { return fps; }
float SceneLoader::getAnimationTimeLength() { return totalAnimationTime; }

